namespace MyRM
{
    public class MyLM : TP.LM
    {
		/*
	  	   Doesn't do lock conversions, so T1(read) -> T1(write) will cause T1 to deadlock with itself.
		*/
		
		public class ResourceLocked : System.Exception
		{
			public ResourceLocked() : base()
			{
			}
			public ResourceLocked( string message ) : base( message )
			{
			}
			public ResourceLocked( string message, System.Exception e  ) : base( message, e  )
			{
			}
		}

		public class DeadLockDetected : System.Exception
		{
			public DeadLockDetected() : base()
			{
			}
			public DeadLockDetected( string message ) : base( message )
			{
			}
			public DeadLockDetected( string message, System.Exception e  ) : base( message, e  )
			{
			}
		}

		public enum LockMode
		{
			Null,
			Read,
			SharedWrite,
			Write,
			Exclusive,
			_Length // the number of LockMode's
		}

		static bool [,] CompatibilityTable = new bool[ (int) MyLM.LockMode._Length, (int)MyLM.LockMode._Length ]
		{
			{	// Null
				true, true, true, true, true	
			},
			{	// Read
				true, true, false, false, false	
			},
			{	// SharedWrite
				true, false, true, false, false	
			},
			{	//	 Write
				true, false, false, false, false	
			},
			{	// Exclusive
				true, false, false, false, false	
			}
		};

		class ResourceEntry
		{
			/*
			  A ResourceEntry is a lock entry in the lock table.  It includes a hash table, 
			  "transactions", for each lock mode, containing transactions holding lock in that mode, 
			  and a field "locked" containing the strongest lock mode held by any transaction on this 
			  resource.
			*/

			public ResourceEntry() // the constructor
			{
				this.transactions = new System.Collections.Hashtable[ (int) MyLM.LockMode._Length];
				this.locked = MyLM.LockMode.Null;
			}

			public bool Compatible( LockMode request ) // can a LockMode lock request be granted?
			{
				return CompatibilityTable[ (int)this.locked, (int)request];
			}

			public void Register( TP.Transaction context, LockMode request )
			{
				// Add a lock of mode "request" for transaction "context"
				// First get the hash table for this lock mode, if it exists
				System.Collections.Hashtable transactionList = this.transactions[(int)request];
				if( transactionList == null )
				{
					// no hash table for this lock mode, so create one
					transactionList = new System.Collections.Hashtable();
					this.transactions[(int)request] = transactionList;
				}
				transactionList[context] = context;
				if( request > locked )      // update the strongest lock mode, if necessary
					locked = request;
				if( evnt != null)           
					evnt.Reset();
			}

			public void Unregister( TP.Transaction context, LockMode request )
			{
				// Release a lock of mode "request" for transaction "context"
				// First get the hash table for this lock mode
				System.Collections.Hashtable transactionList = this.transactions[(int)request];
				if( transactionList == null )
					return; // wasn't registered;
				transactionList.Remove( context );
				for( LockMode l = request; l > LockMode.Null ; locked = --l )
				{
					// recalculate the strongest lock mode
					System.Collections.Hashtable nextTransactionList = this.transactions[(int)l];
					if( nextTransactionList == null )
						continue;
					if( nextTransactionList.Count > 0 )
						break;
				}
				if( request > locked )
					if( evnt != null )
						evnt.Set(); // if anyone was waiting for this lock, they should recheck
			}

			System.Threading.ManualResetEvent evnt;
			public System.Threading.ManualResetEvent UnlockEvent
			{
				get
				{
					lock(this)
					{
						if(evnt == null)
							evnt = new System.Threading.ManualResetEvent(false);
					}
					return evnt;
				}
			}

			System.Collections.Hashtable	[] transactions; // one entry for each LockMode
			LockMode locked;
		}

		public MyLM()
		{
			this.ResourceTable = new System.Collections.Hashtable();
		}

		void Lock( TP.Transaction context, TP.RS resource, LockMode mode )
		{
			ResourceEntry lockTarget;
			lock(this.ResourceTable) // get exclusive access to the lock table
			{
				lockTarget = (ResourceEntry) this.ResourceTable[resource];
				if( lockTarget == null ) // create a ResourceEntry for resource, if there is none
				{
					lockTarget = new ResourceEntry();
					this.ResourceTable[resource] = lockTarget;
				}
			}

			for( int c = 0;; c++)
			{
				if( c > 0 )   // 5 second timeout for deadlock
					if( !lockTarget.UnlockEvent.WaitOne( 5000, false )) 
						throw new DeadLockDetected( string.Format( "Resource {0} timed out" , resource ));
				if( c > 0 )
					System.Console.WriteLine( string.Format( "Attempt {0} in resource {1}", c, resource ));
				lock( lockTarget ) // get exclusive access to the ResourceEntry
				{
					if( lockTarget.Compatible( mode )) // set the lock, if you can
					{
						lockTarget.Register( context, mode );
						return;
					}
				}

			}

			throw new System.Exception( "Internal Error" );
		}

		public void LockForRead( TP.Transaction context, TP.RS resource)
		{
			Lock( context, resource, MyLM.LockMode.Read );	
		}

		public void LockForWrite( TP.Transaction context, TP.RS resource)
		{
			Lock( context, resource, MyLM.LockMode.Write );	
		}

		public void Unlock( TP.Transaction context, TP.RS resource, LockMode mode )
		{
			ResourceEntry lockTarget;
			lock(this.ResourceTable) // get exclusive access to the lock table
			{
				lockTarget = (ResourceEntry) this.ResourceTable[resource];
				if( lockTarget == null )
					return; // wasn't locked
			}
			lock( lockTarget )
			{
				lockTarget.Unregister( context, mode );
			}
			
		}

		public void UnlockRead( TP.Transaction context, TP.RS resource )
		{
			Unlock( context, resource, LockMode.Read );
		}

		public void UnlockWrite( TP.Transaction context, TP.RS resource )
		{
			Unlock( context, resource, LockMode.Write );
		}

		public void UnlockAll( TP.Transaction context )
		{
			// not implemented	
		}

		System.Collections.Hashtable ResourceTable;
    }
}
